#include <stdio.h>
#include <pcap.h>
#include <sys/utsname.h>

#include "vbrsr.h"

#define RT

#define LINE_LENGHT 90
#define HANDLER_TIME_S 10
#define DEFAULT_TIME 5     // In ms
#define DELAY_THR 10       // Soglia % per valutare il tempo di interrarrivo
#define MAC_HDR_DIM 14

#define ESCAPE 27
#define RED "[31m"
#define GREEN "[32m"
#define BLUE "[34m"
#define DEFAULT "[0m"

FILE* file;
pcap_t* handle;  

int seq_num=-1;             // Numero di sequenza del pkt ricevuto
int pkt_lost=0;

double first_time_pkt;     // Per il calcolo del throughput
double last_time_pkt;
int length;

int pkt_lost_line=0;       // Statistica sulla pkt loss sulla linea (istantanea)
float time_factor=0;       // Statistiche sui tempi di ricezione
int out_of_time=0;
int out_of_time_m=0;       // Pkt fuori soglia di tempo lungo tutta la misura
int seq_pkt=0;             // Rilevo almeno due pkt in sequenza : utilizzato per il calcolo del jitter
                           // Numero dei pkt ricevuti in sequenza dopo l'ultima perdita
int seq_pkt_l=0;           // Numero dei pkt sui cui e' stato possibile calcolare il jitter
                           // all'interno della linea;
int seq_pkt_m=0;           // Rilevati in sequenza all'interno di tutta a misura

struct itimerval timer;

void print_jitter(double delay)
{
  double jitter;
  if(seq_pkt < 2)
    fprintf(file,"INF\n");
  else
  {
    jitter=delay-time_factor;
    fprintf(file,"%+f\n",jitter);
  }   
}

void end_rx()
{
 printf("\n\n");
 printf("Statistics   : \n\n");
 printf("Pkts rx      : %i\n",seq_num+1);
 printf("Pkts lost    : %i\n",pkt_lost);
 printf("Pkt loss     : %.3f%%\n",(double)pkt_lost/(double)(seq_num+1)*100);
 printf("Out of time  : %.2f%%\n",(double)out_of_time_m/seq_pkt_m*100);
 printf("Elapsed time : %5.2f s\n",last_time_pkt-first_time_pkt);
 printf("Throughput   : %5.3e bit/s\n",((seq_num+1)-pkt_lost)*8*(length-MAC_HDR_DIM)/(last_time_pkt-first_time_pkt));
 printf("\n");
 exit(1); 
}

void view_rx(char ch)
{
 seq_num++;
 seq_pkt=(ch == 'x') ? 0 : seq_pkt+1;
 seq_pkt_l++;
 seq_pkt_m++;

 if(ch == 'R' || ch == 'A')
 {
   out_of_time++;   
   out_of_time_m++;
 }
 
 // Statistiche di fine riga
 if((seq_num % LINE_LENGHT == 0) && (seq_num != 0))
 { 
  double pli=(double)pkt_lost_line/(double)LINE_LENGHT*100;
  int j =(int)round(pli*13/100);
  printf(" |%6i|%6.2f%%|",seq_num,(double)pkt_lost/(double)seq_num*100);
  for(int i=0;i<j;i++)
    printf("*");
  for(int i=j;i<13;i++)
    printf(" ");
  printf("|");
  pkt_lost_line=0;
  if(ch == 'R' || ch == 'A')
  {
   printf("%5.1f%%|",(double)(out_of_time-1)/seq_pkt_l*100);
   out_of_time=1;
  }
  else
  {
   printf("%5.1f%%|",(double)out_of_time/seq_pkt_l*100);
   out_of_time=0;
  }
  seq_pkt_l=0;
  printf("\n");
 }
 // Stampa a video del carattere equivalente allo stato del frame ricevuto
 switch (ch)
 {
  case 'R':
   printf("%c%s.%c%s",ESCAPE,RED,ESCAPE,DEFAULT);
   break;
  case 'A':
   printf("%c%s.%c%s",ESCAPE,BLUE,ESCAPE,DEFAULT);
   break;
  case 'x':
   printf("x");
   break;
  case '.':
   printf("%c%s.%c%s",ESCAPE,GREEN,ESCAPE,DEFAULT);
   break; 
 }

 fflush(stdout);
}

void print_stat_not_rx(int j)
{
 double var=0;
 if(j == -1)
  fprintf(file,"%f\t\t0\t%i\t0\t0\t0\t0\tINF\n",var,
                                                seq_num);
}

void print_stat_tx(const u_char* pkt)
{
 struct utsname un;
 time_t now=time(NULL);

 uname(&un);

 fprintf(file, "\n# Starting a new log at %s", ctime(&now));
 fprintf(file, "# Sysname  : %s\n",un.sysname);
 fprintf(file, "# Nodename : %s\n",un.nodename);
 fprintf(file, "# Release  : %s\n",un.release);
 fprintf(file, "# Version  : %s\n",un.version);
 fprintf(file, "# Machine  : %s\n",un.machine);
 
 fprintf(file, "# %s is the interface\n", ifname);


 if (get_bit_rate() >= 0)
  fprintf(file, "# %s is the bit rate policy : %iMb\n",(get_auto_bit_rate()) ? "auto" : "fixed",
                                                       get_bit_rate());
 if (get_channel_number() >= 0)
  fprintf(file, "# %d is the used channel\n", get_channel_number());
 if (get_retry_limit(IW_RETRY_MIN) >= 0)
  fprintf(file, "# %d is the short retry limit\n"
	        "# %d is the long retry limit\n",
		get_retry_limit(IW_RETRY_MIN), get_retry_limit(IW_RETRY_MAX));
 iwrange *r;
 if ((r = get_ranges()) != NULL)
   fprintf(file, "# %u is the signal quality range\n"
		 "# %u is the signal level range\n"
		 "# %u is the signal noise range\n",
		r->max_qual.qual, r->max_qual.level, r->max_qual.noise);

 fprintf(file,"\n# MAC Addr RX : ");
 for(int i=0;i<6;i++)
  if(i<5)
   fprintf(file,"%x:",pkt[i]);
  else
   fprintf(file,"%x",pkt[i]); 
 fprintf(file,"\n# MAC Addr TX : "); 
 for(int i=6;i<12;i++)
  if(i<11)
   fprintf(file,"%x:",pkt[i]);
  else
   fprintf(file,"%x",pkt[i]);
 fprintf(file,"\n# Tempo di interrarrivo : %fs\n\n",time_factor);
 
 fprintf(file,"# Time\t\t\tLength\tSeq\tQual\tLevel\tNoise\tStatus\tJitter\n");
}


void search_for_wif(char* dev)
{
  int skfd = iw_sockets_open();
  iwrange range;
 
  if(iw_get_range_info(skfd,"eth1", &range) >= 0)
  {
   strcpy(dev,"eth1");
   return;
  }
  if(iw_get_range_info(skfd,"wlan0", &range) >= 0)
  {
   strcpy(dev,"wlan0");
   return;
  }
  strcpy(dev,"NONE");
  return;
}

void print_help()
{
  char dev[10];
  search_for_wif(dev);
  printf("\nIl ricevitore adotta le seguenti politiche :\n");
  printf(" 1) Ogni pkt ricevuto viene registrato solo se ha il campo type pari a 2004(hex)\n");
  printf(" 2) I pkt non ricevuti vengono registrati per il solo numero di sequenza\n");
  printf(" 3) Il ricevitore esce quando non vengono ricevuti pkts per piu' di %is\n\n",HANDLER_TIME_S);
  printf("\nType : ./receive -i<INTERFACE> -s<FILE_NAME> -m<TIME> -h \n\n");
  printf("-i<INTERFACE> : etichetta associata all'interfaccia che verra' utilizzata per la ricezione.\n");
  printf("                default : viene tentata l'apertura delle interfacce <wlan0> o <eth1>\n");
  printf("                          attuale disponibile : %s\n",dev);
  printf("-s<FILE_NAME> : nome del file in cui vengono raccolti i risultati.\n");
  printf("                default : <NOW>.ris\n");
  printf("-m<TIME>      : tempo di interrarrivo dei pkt fissato al trasmettitore\n");
  printf("              : default : %ims\n",DEFAULT_TIME);
  
  printf("\n\n");
}

void handler(int h)
{
  end_rx();
  pcap_close(handle);
  fclose(file);
}

void got_packet(u_char* args,const struct pcap_pkthdr* header,const u_char* pkt)
{
  double now=((double)header->ts.tv_sec)+((double)header->ts.tv_usec)/1000000;
  // Il timer viene interrotto
  timer.it_value.tv_sec=0;
  setitimer(ITIMER_REAL,&timer,NULL);
  
  // Viene catturato il campo type del pkt
  // Il pkt e' corretto solo se ha come campo type 2004 in hex (3204 int)
  char type[2];
  type[0]=(char)pkt[12];
  type[1]=(char)pkt[13];
  if(((int)type[0] != 32) || ((int)type[1] != 4))
  { 
    //timer.it_value.tv_sec=HANDLER_TIME_S;   
    //setitimer(ITIMER_REAL,&timer,NULL);
   return;
  }
  
  // Viene catturato il numero di sequenza del pkt.
  char seqnum[7];
  for(int i=14;i<21;i++)
   seqnum[i-14]=(char)pkt[i];

  // Viene eseguito solo una volta, per il primo pkt ricevuto
  // Serve ad estrarre delle statitistiche sulla misura
  if(seq_num == -1)
  {
   print_stat_tx(pkt);
   first_time_pkt=((double)header->ts.tv_sec)+((double)header->ts.tv_usec)/1000000;
   length=header->len;  
  }

  // E' stato rilevato un numero fuori sequenza
  if(atoi(seqnum) < seq_num+1)
  {
    printf("\nErrore nella generazione dei numeri di sequenza\n");
    fprintf(file,"# Errore nella generazione dei numeri di sequenza,atteso : %i ,arrivato : %i\n",seq_num,atoi(seqnum));
    fprintf(file,"# ");
    //seq_num=atoi(seqnum);
    return;
  }

  // Se e' presente un fuori sequenza e' andato perso un pkt
  // Viene aggiornato il video , il file e allineato il numero di sequenza
  while((seq_num +1) < atoi(seqnum))
  { 
   // Vengono stampate le statitistiche per il/i pkt non ricevuti nel file e a video
   pkt_lost++;
   pkt_lost_line++;
   view_rx('x');
   print_stat_not_rx(-1);
  }

  double delay=now-last_time_pkt;
  if((delay > time_factor+time_factor*DELAY_THR/100) && (seq_pkt > 1))
    view_rx('R');
  else 
  {
   if((delay < time_factor-time_factor*(int)DELAY_THR/100) && (seq_pkt > 1))
     view_rx('A');
   else
     view_rx('.');
  }
  // Vengono stampate nel file le statistiche sul pkt ricevuto
  // I pkt persi non hanno queste statistiche in quanto le schede DLINK non le forniscono
  //  al contrario delle CNET
  iwstats* p_stats;
  p_stats=get_stats();
  fprintf(file,"%f\t%d\t%i\t%i\t%i\t%i\t1\t",now,
	                                     header->len,
                                             atoi(seqnum),
                              	             (p_stats != NULL) ? p_stats->qual.qual  : 0,
                                             (p_stats != NULL) ? p_stats->qual.level : 0,
	                                     (p_stats != NULL) ? p_stats->qual.noise : 0);
  print_jitter(delay);


  last_time_pkt=now;
  // Viene attivato il timer
  timer.it_value.tv_sec=HANDLER_TIME_S;   
  setitimer(ITIMER_REAL,&timer,NULL);
}

int main(int argc,char** argv)
{
  char file_name[100];
  char dev[10];
  char errbuf[PCAP_ERRBUF_SIZE];

  
  // Default file_name
  time_t now=time(NULL);
  static char statfilename[24];
  strftime(statfilename, 24, "%F.%T.ris", localtime(&now));
  strcpy(file_name,statfilename);

  // Default interface
  // Viene eseguita una ricerca preventiva sulle interfacce 
  //  alla ricerca delle WE
  search_for_wif(dev);

  // Default time intervals
  time_factor=(int)DEFAULT_TIME*0.001;

  // Controllo dei parametri in ingresso
  extern char* optarg;
  char ch;
  while((ch=getopt(argc,argv,"i::m::s::c::h")) != -1)
  {
    switch(ch)
    {
     case 'i':
      strcpy(dev,optarg);
      break;
     case 'm':
      if(sscanf(optarg,"%f",&time_factor) != 1 || time_factor <= 0)
      {
       printf("sscanf(...) error : %s\n",strerror(errno));
       exit(-1);
      }
      break;
     case 's':
      strcpy(file_name,optarg);
      break;
     case 'h':
      print_help();
      exit(1);
    }          
  }
  
  file=fopen(file_name,"w");
  if(file == NULL)
  {
    printf("fopen(...) error : %s\n",strerror(errno));
    exit(-1);
  }

   // Compatibilita' con wapi.c
  ifname=dev;
  init_wireless_api();  
  
  // Codice del comando iwlist <iface> scanning (vedi wapi.c)
  //  estratto da iwlist.c, modificato per :
  //  1) Eseguire la ricerca di reti wireless per un tempo piu' lungo 5s contro
  //     i 250ms standard.
  //  2) Dirigere l'output sul file di descrittore file
  printf("Wait for scanning ...\n");
  print_scan_info(file);
   
  // Timer per la chiusura del programma ricevitore
  bzero(&timer,sizeof(struct itimerval));
  
  struct sigaction action;
  action.sa_handler=handler;
  sigemptyset(&action.sa_mask);
  action.sa_flags=0;
  sigaction(SIGALRM,&action,(struct sigaction*)NULL);
   
  // Abilitazione soft real-time e blocco paginazione
#ifdef RT
  struct sched_param sp;
  sp.sched_priority=sched_get_priority_max(SCHED_FIFO);
  if(sp.sched_priority == -1)
  {
    printf("sched_get_priority_max(...) error : %s\n",strerror(errno));
    exit(-1);
  }
  if(sched_setscheduler(0,SCHED_FIFO,&sp) == -1)
  {
    printf("sched_setscheduler(...) error : %s\n",strerror(errno));
    exit(-1);
  }
  mlockall(MCL_FUTURE);
#endif

  printf("Ready to go ! \n\n");
 
  // Viene ritornato il descrittore per l'interfaccia ifname
  handle=pcap_open_live(ifname,BUFSIZ,1,0,errbuf);
  if(handle == NULL)
  {
    printf("pcap_open_live(...) error : %s\n",errbuf);
    exit(-1);
  }
 
  // Il programma entra in loop continuo 
  // Ad ogni pkt ricevuto viene invocata la got_packet(...)
  if(pcap_loop(handle,-1,got_packet,NULL) == -1)
  {
    printf("pcap_loop(...) error : %s\n",errbuf);
    exit(-1);
  }
}
