/*
 * Copyright (C) 2004 Francesco Potort
 * This program is released under the GNU GPL; see file COPYING for details
 *
 *			S E N D. C
 * This program creates an AF_PACKET socket (Linux) and sends packets
 * on it.  Call it with -h option for usage.
 */

// TODO:

// verificare statistiche del driver in modo ad hoc
// verificare se il maxdelay  giusto (con 2000 pkt sembra di no)
// latency massima comprende tempo invio primo pacchetto?
// confrontare comportamento con e senza hrt
// provare a giocare con la priorit di ksoftirqd
// rapporto risultati in lista hrt

// il datafile non viene riempito pi quando il file finisce!

// fare opzione -u --udp
// fare opzione -V --version
// usare getopt_long
// fare lunghezze variabili
// fare i codici di ritorno giusti

// come fa kismet?
// forse mmap invece di read?


#include "vbrsr.h"

#define BUFS_IN_RING	1024		/* number of buffers in the ring */
#define MAX_READ_BLOCK  (8 * 512)	/* size of max block to read(2) */
#define DUMMY_NO	(HRT ? 4 : 2)	/* # of initial dummy alarm signals */
#define DUMMY_TIME      0.0011		/* length of dummy intervals [s] */

static struct timespec maxlateness;	/* max delay vs. scheduled time */
static int staledata;			/* how many times we sent bad data */

struct buffer
  {
    struct buffer *b_next;	/* pointer to the next buffer in the ring */
    struct timespec b_interval;	/* time to the next timer expiration */
    int datalen;		/* data length: if 0 the catcher has sent it */
    unsigned int seqno;		/* sequence number */
    char *datap;		/* data pointer */
  };

struct t_tab
  {
    struct t_tab *t_next;	/* pointer to the next entry in the ring */
    struct timespec t_interval;	/* time to the next timer expiration */
  };

static struct
  {
    struct buffer *buffer;	/* buffer array */
    char *datap;		/* ptr to the data area */
    struct buffer *cur_rp;	/* ptr to the current read buf in the ring */
    struct buffer *cur_wp;	/* ptr to the current write buf in the ring */
    struct timespec start_time;	/* time of the first timer expiration */
    double dstart_time;		/* same, as a double */
    struct itimerspec next_itimer; /* time of the next timer expiration */
# define next_time next_itimer.it_value
    int late_packets;		/* counter of late packets */
    int buf_count;		/* number of buffers in the ring */
    struct t_tab *timetable;	/* timetable array */
    struct t_tab *cur_tp;	/* ptr to the current time entry in the ring */
  } buf;

static char *progname;			/* pointer to the program name */
static struct sockaddr_storage saddr;	/* socket destination address (max storage) */
static FILE *ttfp;			/* file pointer for the timetable file */
static FILE *sfp;			/* file pointer for the stats file */
static int fd;				/* file descriptor for the input file */
static int sd;				/* socket descriptor */
static int TotpacketNo;			/* total packet number */
static int ifindex;			/* interface index */
static timer_t tid;			/* real-time timer identifier */
static struct timespec res, halfres;	/* clock resolution, half res */
static double dres;			/* clock resolution as a scalar */
static struct timespec dcomp;	   /* delay compensation (see dummy signals) */
static time_t now;			/* current time of day for printing */

static u_int32_t pattern_value;		/* pattern value */
static char pattern_type;		/* pattern type */
static bool numbered;			/* put a seq.no. in the packets */
static int pkts_to_send;		/* number of packets to send */
static int every;			/* generate one error for every bits */
static int pktlen;			/* packet length */
static float time_factor;		/* seconds multiplier */
static bool nosend;			/* don't send packets */
static bool quiet;			/* be quiet */
static bool verbose;			/* verbose messages */
static char srcaddr[3*IFHWADDRLEN];	/* source MAC address */
static char *dstaddr;			/* destination MAC address */

static const inline int min(int a, int b) { if (a < b) return a; else return b; }


static	void
usage (void)
{
  fprintf(stderr, "usage: %s [options] macaddr\n\
	For a description of the options, try %s -h.\n", progname, progname);
  exit(2);
} /* usage() */


static	void
help (void)
{
  fprintf(stderr, "\n\
usage: %s [options] [macaddr]\n\
\n\
 where <macaddr> is destination host IEEE802 MAC address {ff:ff:ff:ff:ff:ff}\n\
\n\
	Defaults are between braces: {...}.  Note that optional arguments\n\
	should follow the option without intervening spaces.  Options are:\n\
\n\
	-D|F|I|R <value> specifies the data used to fill the packets {I}:\n\
		D  data from file (value is file name) {standard input}\n\
      		F  four bytes value {%08x}\n\
		I  value {0} incremented each byte, same for each packet\n\
		R  random data for each packet, value is seed {current time}\n\
	   for the F,I,R flags, if the argument begins with N, that means\n\
	   put a newline-terminated sequence number in the first 8 bytes.\n\
	-c <number of packets> is the number of packets to be sent {1}\n\
	-e <every> error generator: flip one bit every <every> bits\n\
	-i <name> physical interface {loopback}\n\
	-l <lenght> is the packet length in bytes {%u}\n\
	-m <multiplier> multiplies the timetable entries {1}\n\
		default is intergeneration intervals expressed in seconds,\n\
		1e-3 means intervals are expressed in milliseconds\n\
	-n nosend: do not really send packets\n\
	-s [<sfile>] name of stat file to append to {YYYY-MM-DD.HH:MM:SS.vsr}\n\
	-t <timefile> name of timetable file, containing an intergeneration\n\
	   interval per line {a single interval of 1}.  See -m option\n\
	-q quiet: do not print summary at the end\n\
	-v verbose: print waiting time to next packet on stdout\n\
	-h	help: print this page.\n",
	   progname, default_pattern, default_pktlen);
  exit(2);
} /* help() */

/*
 * Allocate stack pages and write to them, then release them on exit.
 */
static void
grow_stack(unsigned int pageno)
{
  char automatic[pageno*PAGE_SIZE];	/* C99 */

  for (int i = 0; i < pageno; i++)
    automatic[i*PAGE_SIZE] = 0x69;	/* force copy on write */
}


/*
 * Open raw socket on which to send messages.
 *
 * input:  destination host name, read from the command line.
 * output: none or error message.
 */
static bool
open_raw_socket (char *dstaddr)
{
  /* create socket on which to send */
  /* The packet(7) man page says that setting the protocol number in
     sll_family is useless when sending, but in fact the third argument
     of the socket syscall is ignored and setting sll_family before
     sendto is necessary. */
  if ((sd = socket(AF_PACKET, SOCK_DGRAM, htons(ETH_P_WICHITA))) < 0)
    { perror("while opening socket"); return FALSE; }

  /* Find the interface index */
  if ((ifindex = if_nametoindex(ifname)) == 0)
    {
	fprintf(stderr, "bad interface %.*s\n", IF_NAMESIZE, ifname);
	usage();			/* does not return */
    }

  /* Find the the source MAC address for printing */
  struct ifreq ifr;
  bzero(&ifr, sizeof(struct ifreq));
  strncpy(ifr.ifr_name, ifname, IF_NAMESIZE);
  if (ioctl(sd, SIOCGIFHWADDR, &ifr) == -1)
    { perror("while reading local MAC address"); return FALSE; }
  for (int i = 0; i < IFHWADDRLEN; i++)
    sprintf(srcaddr+3*i, "%02X:", (unsigned char)ifr.ifr_hwaddr.sa_data[i]);
  srcaddr[17] = '\0';

  /* Create address for sendto */
  struct sockaddr_ll zeroaddr = {0};
  struct sockaddr_ll *sap = (struct sockaddr_ll *)&saddr;
  *sap = zeroaddr;			/* clear the address struct */
  sap->sll_family = AF_PACKET;		/* address family */
  /* The packet(7) man page says that setting the protocol number is
     useless when sending, but in fact the third argument of the socket
     syscall is ignored and setting this is necessary. */
 sap->sll_protocol = htons(ETH_P_WICHITA); /* our private protocol number */

  for (int i = 0; i < IFHWADDRLEN; i++)	/* parse the destination MAC address */
    {
      int addrbyte = 0;
      if (dstaddr != NULL)		/* only do it if there is an address */
	{
	  //dstaddr[3*i] &= ~('a' - 'A');	/* upcase for esthetic reasons */
	  //dstaddr[3*i+1] &= ~('a' - 'A');
	  if (sscanf(dstaddr+3*i, "%2x", &addrbyte) != 1
	      || (i < IFHWADDRLEN-1 && dstaddr[3*i+2] != ':')
	      || (i == IFHWADDRLEN-1 && dstaddr[3*i+2] != '\0'))
	    { fprintf(stderr, "bad Ethernet MAC address\n"); return FALSE; }
	}
      sap->sll_addr[i] = addrbyte;	/* write the address in the sockaddr_ll */
    }
  sap->sll_halen = IFHWADDRLEN;		/* address length */
  sap->sll_ifindex = ifindex;		/* interface index */

  return TRUE;
} /* open_raw_socket() */

/*
 * Generate error for every bits.
 *
 * input:  Buffer pointer.
 * output: none.
 */
int target_packet, target_byte, target_bit;
int target_count, error_bit, byte;
static void
error_generator (struct buffer *bufp)
{
  for (target_byte = target_byte % pktlen;
       target_byte < pktlen;
       target_byte += byte)
    bufp->datap[target_byte] ^= (1 << (target_bit));

  if (target_byte < pktlen)
    target_byte += byte;

} /* error_generator() */


/*
 * Open timetable input file.
 *
 * input:  input file name
 * output: none or error message.
 */
static bool
read_timefile (char *tfilename, float time_factor)
{
  /* Case of no timetable file. */
  if (tfilename == NULL)
    {
      static struct t_tab tt;

      tt.t_next = &tt;
      loadtspec(&tt.t_interval, NSEC_PER_SEC * time_factor * 1 + 0.5);
      buf.timetable = &tt;

      return TRUE;
    }

  /* Open input file. */
  if (!(ttfp = fopen(tfilename, "r")))
    { perror("while opening timetable file"); return FALSE; }

  /* Read timetable entry from the input file. */
  int lineno = 1;			/* line number */
  int t_entries = BUFSIZ;
  buf.timetable = malloc(t_entries * sizeof(struct t_tab));

  while (TRUE)
    {
      if (buf.timetable == NULL)
        {
          fprintf(stderr, "not enough memory to allocate timetable\n");
          return FALSE;
        }

      for (struct t_tab *tp = buf.timetable + t_entries - BUFSIZ;
	   tp < buf.timetable + t_entries;
	   tp++)
        {
	  float t_interval;

          /* Time in seconds*time_factor. */
          if (fscanf(ttfp, "%f\n", &t_interval) != 1)
            {
              if (feof(ttfp))
		{
		  /* Link the ring. */
		  tp--;
		  tp->t_next = buf.timetable;
		  for (tp--; tp >= buf.timetable; tp--)
		    tp->t_next = tp+1;
		  return TRUE;
		}
              fprintf(stderr,
		      "error reading timetable file at line %d.\n", lineno);
              return FALSE;
            }

	  lineno += 1;

          /* Time in nanoseconds. */
          loadtspec(&tp->t_interval,
		    NSEC_PER_SEC * time_factor * t_interval + 0.5);
        }

      /* Allocate more space for the timetable entries. */
      t_entries += BUFSIZ;
      buf.timetable = realloc(buf.timetable,
			      t_entries * sizeof(struct t_tab));
    } /* forever */
} /* read_timefile() */


/*
 * Make `buffer' ring with `pattern' data.
 *
 * input:  Fixed, incremental or random pattern_type and pattern value.
 * output: none or error message.
 */
static bool
fill_ring_with_pattern (void)
{
  register struct buffer *bufp;		/* buffer pointer */
  register struct t_tab *timep;		/* timetable pointer */

  /* allocate data space */
  buf.datap = calloc(BUFS_IN_RING, pktlen);
  if (buf.datap== NULL)
    {
      fprintf(stderr, "not enough memory to allocate data space\n");
      return FALSE;
    }

  /* allocate buffer space */
  buf.buffer = calloc(BUFS_IN_RING, sizeof(struct buffer));
  if (buf.buffer == NULL)
    {
      fprintf(stderr, "not enough memory to allocate buffer space\n");
      return FALSE;
    }

  /* Create the data and put it in the ring. */
  int ready_bufs;
  bufp = NULL;				/* make the compiler quiet */
  for (ready_bufs = 0, timep = buf.timetable;
       ready_bufs < min(BUFS_IN_RING, pkts_to_send);
       ready_bufs++, timep = timep->t_next)
    {
      bufp = &buf.buffer[ready_bufs];
      bufp->datap = &buf.datap[ready_bufs*pktlen];
      bufp->datalen = pktlen;

      /* fill sending buffer with pattern */
      switch (pattern_type)
        {
        case FIXED:		/* fixed pattern */
          for (int i = 0; i < pktlen/sizeof(int); i++)
	    ((u_int32_t *)bufp->datap)[i] = pattern_value;
        break;

        case INCREMENTAL:	/* incremental pattern */
          for (int i = 0; i < pktlen; i++)
	    bufp->datap[i] = i + pattern_value;
        break;

	case RANDOM:		/* random pattern */
          /* random starting point */
          srand(pattern_value);
	  for (int i = 0; i < pktlen/sizeof(short); i++)
	    ((short *)bufp->datap)[i]  = rand();
        break;
	} /* switch(pattern_type) */

      /* bit_error generator */
      if (every != 0)
	{
	  if (target_count >= target_packet)
	    {
              error_generator(bufp);
	      target_count = 0;
	    }
	  else target_count++;
	}

      /* Set packet sequence number. */
      if (numbered && pktlen >= 8)
	{
	  bufp->seqno = ready_bufs;
	  sprintf(bufp->datap, "%07u\n", bufp->seqno);
	}

      /* time in nanoseconds */
      bufp->b_interval = timep->t_interval;
      bufp->b_next = bufp + 1;
    }
  bufp->b_next = buf.buffer;

  buf.cur_tp = timep->t_next;
  buf.cur_wp = buf.cur_rp = buf.buffer;

  return TRUE;
} /* fill_ring_with_pattern() */


/*
 * Open input file and make `buffer' ring.
 *
 * input:  input file name
 * output: none or error message.
 */
static bool
fill_ring_from_datafile (char *dfilename)
{
  register struct buffer *bufp;		/* buffer pointer */
  struct t_tab *timep;
  int nread;
  int bufs_to_ready = min(pkts_to_send, BUFS_IN_RING);

  /* Allocate data space. */
  buf.datap = calloc(bufs_to_ready, pktlen);
  if (buf.datap == NULL)
    {
      fprintf(stderr, "not enough memory to allocate data space\n");
      return FALSE;
    }

  /* Allocate buffer space. */
  buf.buffer = calloc(bufs_to_ready, sizeof(struct buffer));
  if (buf.buffer == NULL)
    {
      fprintf(stderr, "not enough memory to allocate buffer space\n");
      return FALSE;
    }

  /* Open data file. */
  if (dfilename == default_datafile)
    fd = STDIN_FILENO;
  else if ((fd = open(dfilename, O_RDONLY)) < 0)
    { perror("while opening data file"); return FALSE; }

  /* Read the data and put it in the ring. */
  if ((nread = read(fd, buf.datap, bufs_to_ready * pktlen)) < 0)
    { perror("while reading data file"); return FALSE; }
  if (nread == 0)
    { fprintf(stderr, "file %s is empty\n", dfilename); return FALSE; }

  bufp = NULL;				/* make the compiler quiet */
  if (nread <= pktlen)
    {
      bufp = &buf.buffer[0];
      bufp->datap = &buf.datap[0];
      bufp->datalen = nread;
      timep = buf.timetable;
      bufp->b_interval = timep->t_interval;
    }
  else
    {
      int ready_bufs;
      for (ready_bufs = 0, timep = buf.timetable;
	   ((nread >= pktlen && ready_bufs < bufs_to_ready)
	    || (nread < pktlen && ready_bufs != 0));
	   nread -= pktlen, ready_bufs++, timep = timep->t_next)
        {
          if (nread <= 0)
	    break;
          bufp = &buf.buffer[ready_bufs];
          bufp->datap = &buf.datap[ready_bufs*pktlen];
      	  if (nread >= pktlen)
            bufp->datalen = pktlen;
          else
            bufp->datalen = nread;
          bufp->b_interval = timep->t_interval;
      	  bufp->b_next = bufp + 1;
    	}
    }
  bufp->b_next = buf.buffer;

  buf.cur_tp = timep;
  buf.cur_wp = buf.cur_rp = buf.buffer;

  return TRUE;
} /* fill_ring_from_datafile() */


/*
 * Monitor the buffer, and refill it as soon as a buffer is sent
 */
static void
keep_buffer_full (char *dfilename)
{
  int ready_bufs = min(pkts_to_send, BUFS_IN_RING);

  /* Data from internal generator. */
  if (dfilename == NULL)
    {
      /* while there are buffers to be filled */
      while (ready_bufs < pkts_to_send)
        {
	  pause();

	  register struct buffer *bufp = buf.cur_rp; /* buffer pointer */
          while (bufp->datalen == 0 && ready_bufs++ < pkts_to_send)
     	    {
	      /* get current timetable entry */
              bufp->b_interval = buf.cur_tp->t_interval;
	      bufp->datalen = pktlen;

	      /* Fill sending buffer. */
              switch (pattern_type)
		{
		case FIXED:		/* fixed pattern */
		  break;
		case INCREMENTAL:	/* incremental pattern */
		  break;
		case RANDOM:		/* random pattern */
          	  /* random starting point */
          	  srand(pattern_value);
		  for (int i = 0; i < pktlen/2; i++)
		    ((short *)bufp->datap)[i] = rand();
		  break;
		} /* switch(pattern) */

	      /* bit_error generator */
      	      if (every != 0)
		{
	  	  if (target_count >= target_packet)
	    	    {
            	      error_generator(bufp);
	    	      target_count = 0;
	    	    }
	  	  else target_count++;
		}

	      /* Set packet sequence number. */
	      if (numbered && pktlen >= 8)
		{
		  bufp->seqno = ready_bufs;
		  sprintf(bufp->datap, "%07u\n", bufp->seqno);
		}

	      buf.cur_rp = bufp = bufp->b_next;
	      buf.cur_tp = buf.cur_tp->t_next;
	    }
	} /* until we send pkts_to_send buffers */
    } /* data from internal generator */

  /* Data from file. */
  else
    {
      int nread;		/* number of bytes read with read(2) */
      int data_to_read;		/* number of bytes to read now */
      struct buffer *bp;
      struct t_tab *tp;
      int threshold = MAX_READ_BLOCK - pktlen;
      bool eof = FALSE;


      /* Read while not EOF */
      do
	{
	  while (buf.cur_rp->datalen > 0)
	    pause();

	  /* save in bp the pointer to the first empty buffer */
	  bp = buf.cur_rp;
	  tp = buf.cur_tp;

	  /* compute the number of the first contiguous empty buffers */
	  data_to_read = 0;
	  do
	    {
	      data_to_read += pktlen;
	      buf.cur_rp = buf.cur_rp->b_next;
	      buf.cur_tp = buf.cur_tp->t_next;
	    }
	  while (buf.cur_rp > buf.buffer
		 && buf.cur_rp->datalen == 0
		 && data_to_read <= threshold);

	  do
	    nread = read(fd, bp->datap, data_to_read);
	  while (nread < 0 && errno == EINTR);

	  if (nread < 0)
	    perrexit2("while reading data file");
	  eof = (nread == 0);


	  /* buffers are now filled */
	  for (; nread > 0; nread -= pktlen)
	    {
	      bp->datalen = pktlen;
	      bp->b_interval = tp->t_interval;
	      bp = bp->b_next;
	      tp = tp->t_next;
	    }
	}
      while (!eof);

    } /* read from the input file */
} /* keep_buffer_full() */

static void
print_headers (void)
{
  if (verbose)				/* verbose flag */
    {
      iwrange *r;

      printf("  -Interface is %s%s",
	     ifname, (nosend) ? " (nosend)" : "");
      printf("  -MAC addresses (src, dst): %s, %s",
	     srcaddr, (nosend) ? "(nosend)" : dstaddr);
      if (get_bit_rate() >= 0)
	printf("\n  -Bit rate is %s",
	       (get_auto_bit_rate()) ? "automatic" : "fixed");
      if (get_channel_number() >= 0)
	printf("  -Channel: %d", get_channel_number());
      if (get_retry_limit(IW_RETRY_MIN) >= 0)
	printf("  -Retry limits (short, long): %d, %d",
	       get_retry_limit(IW_RETRY_MIN), get_retry_limit(IW_RETRY_MAX));
      if ((r = get_ranges()) != NULL)
	printf("\n  -Ranges for signal quality, level, noise: %u, %u, %u",
	       r->max_qual.qual, r->max_qual.level, r->max_qual.noise);
    }


  if (sfp)				/* stats file */
    {
      fprintf(sfp, "#\n#-- Starting a new log at %s", ctime(&now));
      fprintf(sfp, "# %s is the interface%s\n", ifname,
	      (nosend) ? " (nosend)" : "");
      fprintf(sfp, "# %s is the source MAC address\n", srcaddr);
      fprintf(sfp, "# %s is the destination MAC address\n",
	     (nosend) ? "(nosend)" : dstaddr);
      if (get_bit_rate() >= 0)
	fprintf(sfp, "# %s is the bit rate policy\n",
		(get_auto_bit_rate()) ? "auto" : "fixed");
      if (get_channel_number() >= 0)
	fprintf(sfp, "# %d is the used channel\n", get_channel_number());
      if (get_retry_limit(IW_RETRY_MIN) >= 0)
	fprintf(sfp, "# %d is the short retry limit\n"
		"# %d is the long retry limit\n",
		get_retry_limit(IW_RETRY_MIN), get_retry_limit(IW_RETRY_MAX));
      iwrange *r;
      if ((r = get_ranges()) != NULL)
	fprintf(sfp, "# %u is the signal quality range\n"
		"# %u is the signal level range\n"
		"# %u is the signal noise range\n",
		r->max_qual.qual, r->max_qual.level, r->max_qual.noise);
    }
} /* print_headers() */


static void
print_stats (struct timespec *sigtimep, struct timespec *currentp, bool late_pkt)
{
  static double dprevcurrent, dprevinterval;
  double dcurrent = dtspec(currentp);
  double dinterval = dtspec(&buf.cur_wp->b_interval);
  double dnexttime = dtspec(&buf.next_time);
  iwstats *s = NULL;

  if (TotpacketNo == 1)
    dprevcurrent = dcurrent; /* prevcurrent makes no sense for first packet */

  if (verbose)
    /* Printing timing data on standard output. */
    {
      if (s == NULL) s = get_stats();
      if (TotpacketNo == 1)
	printf("\n  -Reference time is %.4f\n"
"seq   |rate|qua|sig|no |sending  |current   |jitter|scheduled |next      |sending\n"
"number|Mb/s|lty|nal|ise|time [us]|time [s]  |  [us]|intrval[s]|time [s]  |      ?\n",
	       buf.dstart_time);
      printf
	("%6d%5d%4u%4u%4u%10.0f%11.4f%7.0f",
	 TotpacketNo,			/* seq number */
	 get_bit_rate(),		/* bit rate */
	 (s) ? s->qual.qual : 0,	/* raw signal quality */
	 (s) ? s->qual.level : 0,	/* raw signal level */
	 (s) ? s->qual.noise : 0,	/* raw noise level */
	 1e6 * (dcurrent - dtspec(sigtimep)), /* sending time */
	 dcurrent - buf.dstart_time,	/* current time */
	 1e6 * (dcurrent - dprevcurrent - dprevinterval)); /* jitter */
      if (TotpacketNo < pkts_to_send)
	printf((late_pkt) ? "%11.4f%11.4f,    now\n"
	                  : "%11.4f%11.4f,  later\n",
	       dinterval,		/* scheduled interval */
	       dnexttime - buf.dstart_time); /* next time */
      else
	puts("");			/* print newline */
    }

  if (sfp)
    /* Print timing data on stats file. */
    {
      static double dprevnexttime;

      if (s == NULL) s = get_stats();
      if (TotpacketNo == 1)
	{
	  dprevnexttime = buf.dstart_time; /* req. time of first packet */
	  fprintf(sfp, "# %.4f is the reference time\n"
		  "#packet,rate,qua,sig,no ,len.,requested,actual   ,jitter\n"
		  "#number,Mb/s,lty,nal,ise, [B],time  [s],time  [s],  [us]\n",
		  buf.dstart_time);
	}
      fprintf
	(sfp, "%7u,%4d,%3u,%3u,%3u,%4u,%9.4f,%9.4f,%7.0f\n",
	 TotpacketNo,			/* packet number */
	 get_bit_rate(),		/* bit rate */
	 (s) ? s->qual.qual : 0,	/* raw signal quality */
	 (s) ? s->qual.level : 0,	/* raw signal level */
	 (s) ? s->qual.noise : 0,	/* raw noise level */
	 buf.cur_wp->datalen,		/* packet length */
	 dprevnexttime - buf.dstart_time, /* requested time */
	 dcurrent - buf.dstart_time,	/* actual time */
	 1e6 * (dcurrent - dprevcurrent - dprevinterval)); /* jitter */
      dprevnexttime = dnexttime;
    }

  dprevcurrent = dcurrent;
  dprevinterval = dinterval;
} /* print_stats() */


static void
catcher (int sig)
{
  struct timespec sigtime, current;
  bool late_pkt = FALSE;
  bool empty_ring;

  /* Generate some dummy signals at the beginning, in order to compensate for
     the delay of the first packets.  The subsequent packets all exhibit the
     same delay, for unknown reasons, but the jitter is low after the few
     packets sent.  This happens even when using the -n option.  To look at
     this effect, just set dummy_alarms to 0 and use the -v option. */
  static int dummy_alarms = DUMMY_NO;
  if (dummy_alarms > 0)
    {
      /* Dummy signals are generated about every tenth of second. */
      static struct timespec dummy_interval = {0, NSEC_PER_SEC * DUMMY_TIME};

      /* Compute new sending time. */
      addtspec(&buf.next_time, &dummy_interval);

      /* Set alarm. */
      if (timer_settime(tid, TIMER_ABSTIME, &buf.next_itimer, NULL) < 0)
	perrexit2("cannot set the timer");

      /* One dummy alarm set. */
      dummy_alarms -= 1;

      return;
    }

  /* Get time of signal for statistical or initialisation purposes. */
  if (verbose || sfp			/* statistics */
      || dummy_alarms == 0)		/* initialisation */
    {
      /* Get the current time. */
      if (clock_gettime(CLOCKID, &sigtime) != 0)
	perrexit2("cannot read signal time");
    }

  /* Init the time base. */
  if (dummy_alarms == 0)		/* this happens only once */
    {
      if (!quiet)
	{
	  /* Compute and store the initial delay compensation. */
	  dcomp = sigtime;		/* store the current time */
	  subtspec(&dcomp, &buf.next_time); /* subtract the requested time */
	}

      buf.start_time = sigtime;		/* current time is origin of times */
      buf.dstart_time = dtspec(&buf.start_time); /* save it as a scalar */
      dummy_alarms = -1;		/* initialisation completed */
    }

  do
    {
      /* Check if the data buffer we are going to send has been set up
	 in the main synchronous loop.  If datalen is 0, that means we already
	 sent it, but it hasn't been refilled yet: the data in it are is old
	 (stale) one, and so is the time interval. */
      empty_ring = (buf.cur_wp->datalen == 0);
      if (empty_ring)
	{
	  staledata += 1;

	  /* We do not resend the old stale data, and set the next interval to
	     1 second. */
	  loadtspec(&buf.next_time, 1);
	  break;
	}

      /* Lateness is how much a packet is late with respect to the scheduled
	 sending time.  We use the signal time or, if this is a late packet,
	 the time after the sending of the previous packet.
         Notice that buf.next_time is the time when the current packet should
	 have been sent: we have not yet advanced it. */
      struct timespec lateness = (late_pkt) ? current : sigtime;
      subtspec(&lateness, &buf.next_time); /* compute packet lateness */
      if (gttspec(&lateness, &maxlateness))
	maxlateness = lateness;		/* compute max lateness */

      if (!nosend)
	/* Send message. */
	if (sendto(sd, buf.cur_wp->datap, buf.cur_wp->datalen, 0,
		   (struct sockaddr *)&saddr, sizeof(struct sockaddr_ll)) < 0)
	  perrexit2("sendto");

      /* Get current time. */
      if (clock_gettime(CLOCKID, &current) != 0)
	perrexit2("cannot read current time");

      /* Increment total packet number. */
      TotpacketNo++;

      /* Compute new sending time. */
      if (TotpacketNo < pkts_to_send)
	{
	  addtspec(&buf.next_time, &buf.cur_wp->b_interval);

	  /* Has the time expired? */
	  struct timespec threshold = current;
	  addtspec(&threshold, &halfres); /* now plus (clock resolution)/2 */
	  late_pkt = (!gttspec(&buf.next_time, &threshold));

	  if (late_pkt)
	    /* The time when the next packet should have been sent has already
	       expired (or is very close to expire, less than half the timer
	       resolution).  We send it immediately. */
	    {

	      /* If the current time is greater than the sending time, we count
		 this as a late packet.  If we are sending it because the
		 expiration time is near, but has not yet expired, we don't
		 increment the late packet counter. */
	      if (gttspec(&current, &buf.next_time)) /* later than next? */
		buf.late_packets++;
	    }

	  /* Print statistics. */
	  if (verbose || sfp)
	    print_stats(&sigtime, &current, late_pkt);
	}
      else
	{
	  /* Print statistics. */
	  if (verbose || sfp)
	    print_stats(&sigtime, &current, FALSE);

	  return;			/* no more packets to send */
	}

      /* More to send: mark this one as already sent. */
      buf.cur_wp->datalen = 0;
      buf.cur_wp = buf.cur_wp->b_next;

    } while (late_pkt);

  /* Set alarm. */
  if (timer_settime(tid, TIMER_ABSTIME, &buf.next_itimer, NULL) < 0)
    perrexit2("cannot set the timer");

  return;
} /* catcher() */

int
main (int argc, char **argv)
{
  char c;
  extern char *optarg;
  extern int optind;
  char *dfilename = NULL;		/* input data file name */
  char *sfilename = NULL;		/* output stats file name */
  char *tfilename = NULL;		/* timetable file name */
  struct timespec stop_time;
  struct sched_param prio;


  /* Initialize variables */
  loadtspec(&buf.next_itimer.it_interval, 0);
  now = time(NULL);
  progname = argv[0];

  /* Read options */
  while ((c = getopt(argc, argv,
		     "D::F::I::R::be:i:l:m:nc:qs::t:vh")) != -1)
    switch (c)
      {
      case DATAFILE:
      case FIXED:
      case INCREMENTAL:
      case RANDOM:
	if (pattern_type != '\0') usage(); /* already assigned */
	pattern_type = c;
	if (optarg != NULL && optarg[0] == 'N')
	  {
	    numbered = TRUE;		/* first char N means number pkts */
	    if (optarg[1] == '\0')
	      optarg = NULL;		/* nothing else */
	    else
	      optarg += 1;		/* something else is there */
	  }
	if (optarg == NULL)
	  switch (c)
	  {
	  case DATAFILE:
	    dfilename = default_datafile;
	    break;
	  case FIXED:
	    pattern_value = default_pattern;
	    break;
	  case INCREMENTAL:
	    pattern_value = 0;
	    break;
	  case RANDOM:
	    pattern_value = time(NULL);
	    break;
	  }
	else switch (c)
	  {
	  case DATAFILE:
	    dfilename = default_datafile;
	    break;
	  case FIXED:
	  case INCREMENTAL:
	  case RANDOM:
	    if (sscanf(optarg, "%i", &pattern_value) != 1
		|| pattern_value <= 0)
	      {
		fprintf(stderr, "argument to option -%c must be positive\n", c);
		usage();
	      }
	    break;
	  }
	break;

      case 'c':
	if (pkts_to_send != 0) usage(); /* already assigned */
	if (sscanf(optarg, "%d", &pkts_to_send) != 1
	    || pkts_to_send <= 0)
	  {
	    fprintf(stderr, "number of packets must be positive\n");
	    usage();
	  }
	break;

      case 'e':
	if (every != 0) usage();	/* already assigned */
	if (sscanf(optarg, "%d", &error_bit) != 1
	   || every <= 0)
	  {
	    fprintf(stderr, "bit error position must be positive\n");
	    usage();
	  }
	target_packet = error_bit / (pktlen * 8);
	target_byte = error_bit / 8;
	byte = target_byte;
	target_bit = error_bit % 8;
	break;

      case 'i':
	if (ifname != NULL) usage();	/* already assigned */
	ifname = optarg;
	break;

      case 'l':
	if (pktlen != 0) usage();	/* already assigned */
	if (sscanf(optarg, "%d", &pktlen) != 1
	    || pktlen <= 0)
	  {
	    fprintf(stderr, "packet length must be positive\n");
	    usage();
	  }
	break;

      case 'm':
	if (time_factor != 0) usage();	/* already assigned */
	if (sscanf(optarg, "%f", &time_factor) != 1
	    || time_factor <= 0)
	  {
	    fprintf(stderr, "multiplier must be positive\n");
	    usage();
	  }
	break;

      case 'n':
	if (nosend) usage();		/* already assigned */
	nosend = TRUE;
	break;

      case 's':
	if (sfilename != NULL) usage(); /* already assigned */
	if (optarg == NULL)
	  {
	    /* Create a file name representing the current date usign the format
	       YYYY-MM-DD.hh:mm:ss.vsr */
	    static char statfilename[24];
	    strftime(statfilename, 24, "%F.%T.vsr", localtime(&now));
	    sfilename = statfilename;
	  }
	else
	  {
	    sfilename = optarg;
	  }
	break;

      case 't':
	if (tfilename != NULL) usage(); /* already assigned */
	tfilename = optarg;
	break;

      case 'q':
	if (quiet) usage();		/* already assigned */
	quiet = TRUE;
	break;

      case 'v':
	if (verbose) usage();		/* already assigned */
	verbose = TRUE;
	break;

      case 'h':
	help();
	break;
      case '?':
	usage();
	break;
      } /* while getopt() switch */

  /* Defaults */
  if (pattern_type == 0)
    pattern_type = INCREMENTAL;
  if (pkts_to_send == 0)
    pkts_to_send = default_pkts_to_send;
  if (pktlen == 0)
    pktlen = default_pkts_to_send;
  if (time_factor == 0)
    time_factor = default_time_factor;

  /* Must have at most one argument. */
  if (argc > optind+1)
    usage();

  /* Open timetable file. */
  if (!read_timefile(tfilename, time_factor))
    usage();

  /* Open data file and make `buffer' ring. */
  if (dfilename == NULL)		/* pattern data must be sent */
    {
      if (!fill_ring_with_pattern())
        usage();
    }
  else					/* data send from the input file */
    {
      if (!fill_ring_from_datafile(dfilename))
        usage();
    }

  /* Open stats file. */
  if (sfilename != NULL)
    if ((sfp = fopen(sfilename, "write")) == NULL)
      perrexit2("cannot open the stats file");

  /* Open raw socket. */
  if (optind == argc)			/* if no arguments */
    {
      if (nosend)			/* if -n option used */
	dstaddr = NULL;			/* nothing will be sent */
      else
	dstaddr = "ff:ff:ff:ff:ff:ff";	/* default is broadcast address */
    }
  else					/* argument is MAC address */
    {
      dstaddr = argv[optind];
    }
  if (!open_raw_socket(dstaddr))
    usage();

  /* Init the structures for accessing driver parameters. */
  init_wireless_api();

  /* Print verbose and stats headings. */
  print_headers();

  /* Set real-time scheduling priority and nice value. */
  if (nice(-20) < 0 && errno != 0)
    perrexit2("cannot set nice value");
  prio.sched_priority =
    (sched_get_priority_min(POLICY) + sched_get_priority_max(POLICY)) / 2;
  if ((sched_setscheduler(0, POLICY, &prio) < 0))
    perrexit2("cannot set scheduler policy");
  errno = 0;

  /* Create the timer and get its resolution. */
  if (timer_create(CLOCKID, NULL, &tid) != 0)
    perrexit2("cannot create the timer");
  if (clock_getres (CLOCKID, &res) != 0)
    perrexit2("cannot read the timer resolution");
  dres = dtspec(&res);
  loadtspec(&halfres, dres/2 * NSEC_PER_SEC); /* half resolution */

  /* Set the signal handler. */
  struct sigaction action;
  action.sa_handler = catcher;
  action.sa_flags = 0;
  if (sigemptyset(&action.sa_mask) != 0)
    perrexit2("cannot reset signals");
  (void) sigaction(SIGALRM, &action, NULL);

  /* Get the timer clock. */
  if (clock_gettime(CLOCKID, &buf.next_time) != 0)
    perrexit2("cannot read timer");

  /* Call the signal handler for the first time. */
  catcher(0);

  /* Avoid page faults. */
  grow_stack(4);			/* allocate stack memory pages */
  mlockall(MCL_FUTURE);			/* lock memory pages */

  /* Monitor the buffer and fill it as soon as a packet is sent. */
  if (pkts_to_send > BUFS_IN_RING)
    keep_buffer_full(dfilename);

  /* Check the buffer ring until every packet is sent. */
  while (TotpacketNo < pkts_to_send)
    pause();

  /* Print final summary */
  if (!quiet)
    {
      clock_gettime(CLOCKID, &stop_time);
      printf("  Total run time %g s"
	     " (clock resolution %g ms, delay compensation %g ms)\n",
	     dtspec(&stop_time) - buf.dstart_time,
	     1e3 * dres,
	     1e3 * dres * round(dtspec(&dcomp) / dres));
      printf("  %d packets sent,"
	     "  %d late packets, maximum lateness %g ms\n",
	     TotpacketNo, buf.late_packets,
	     /* We subtract the compensation time from the maximum lateness. */
	     1e3 * dres * round((dtspec(&maxlateness)-dtspec(&dcomp)) / dres));
    }
  if (staledata > 0)
    printf(" ERRORS FOUND: %d times stale data found and pkt not sent\n",
	   staledata);

  return 0;
} /* main () */



/*
 * Local Variables:
 * comment-column: 40
 * fill-column: 79
 * c-font-lock-extra-types: ("FILE"  "\\sw+_t" "bool" "iwrange" "iwstats")
 * End:
 */
