#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <inttypes.h>
#include <errno.h>
#include <sys/user.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>

#include <stdio.h>
#include <signal.h>
#include <string.h>

#define  __USE_XOPEN2K			/* for the real-time clocks */
#include <time.h>			/* struct timespec, itemerspec */
#include <sched.h>			/* sched_setscheduler() */
#include <sys/mman.h>			/* mlockall() */

#include <sys/socket.h>
#include <netinet/in.h>			/* struct in_addr, htons */
#include <netpacket/packet.h>		/* the socket packet interface */
#include <net/ethernet.h>		/* Ethernet protocol numbers */
#include <netinet/ip.h>			/* the IP header */
/* <net/if.h> conflicts with linux/if.h, which is included by iwlib.h */
#define broken_iwlib_h 1
#if broken_iwlib_h
# define IF_NAMESIZE IFNAMSIZ
  extern unsigned int if_nametoindex (__const char *__ifname);
#else
# include <net/if.h>			/* if_nametoindex() */
#endif
#include <iwlib.h>			/* wireless extensions */

#define FALSE	0
#define TRUE	1

#ifdef DEBUG
# define DEBUG TRUE
#else
# define DEBUG FALSE
#endif

#ifndef NSEC_PER_SEC
# define NSEC_PER_SEC (1000000000L)
#endif
#ifndef CLOCK_MONOTONIC_HR
# define CLOCK_REALTIME_HR	4
# define CLOCK_MONOTONIC_HR	5
#endif

#define ETH_P_WICHITA	0x2004	/* Ethernet type for benchmarks */
				/* pattern types */
#define DATAFILE	'D'
#define FIXED		'F'
#define INCREMENTAL	'I'
#define RANDOM		'R'

#define POLICY		SCHED_FIFO	/* scheduling priority */
/* Choose between high-resolution monotonic timer
   and low-resolution wall clock timer. */
#define CLOCKID		(HRT ? CLOCK_MONOTONIC_HR : CLOCK_REALTIME)

#define default_datafile "-"		/* default data file means stdin */
#define default_time_factor 1.0		/* default time factor*/
#define default_pattern 0xF0CACC1A	/* default pattern */
#define default_pktlen 1000		/* default packet length */
#define default_pkts_to_send 1		/* default number of packets to send */
#define default_ifindex 2		/* default interface index */


typedef int bool;

#define perrexit2(string) { perror(string); exit(2); }

/* addtspec (tvdestp, tcsrcp) adds the timespec pointed by tcsrcp
   to that pointed by tvdestp and puts the result in tvdestp. */
#define addtspec(tvdestp,tvsrcp) do { \
  (tvdestp)->tv_sec += (tvsrcp)->tv_sec;		\
  (tvdestp)->tv_nsec += (tvsrcp)->tv_nsec;		\
  if ((tvdestp)->tv_nsec > NSEC_PER_SEC) {		\
      (tvdestp)->tv_sec += 1;				\
      (tvdestp)->tv_nsec -= NSEC_PER_SEC;		\
  }							\
} while (0)

/* gttspec (tvdestp, tcsrcp) returns a non null value if the timespec
   pointed by tvdestp is greater than that pointed by tcsrcp. */
#define gttspec(tvdestp,tvsrcp) \
  ((tvdestp)->tv_sec > (tvsrcp)->tv_sec			\
   || ((tvdestp)->tv_sec == (tvsrcp)->tv_sec		\
       && (tvdestp)->tv_nsec > (tvsrcp)->tv_nsec))

/* subtspec (tvdestp, tcsrcp) subtracts the timespec pointed by tcsrcp
   from that pointed by tvdestp and puts the result in tvdestp. */
#define subtspec(tvdestp,tvsrcp) do { \
  (tvdestp)->tv_sec -= (tvsrcp)->tv_sec;		\
  (tvdestp)->tv_nsec -= (tvsrcp)->tv_nsec;		\
  if ((tvdestp)->tv_nsec < 0) {				\
      (tvdestp)->tv_sec -= 1;				\
      (tvdestp)->tv_nsec += NSEC_PER_SEC;		\
  }							\
} while (0)

/* loadtspec (tvdestp, ns) loads the integer ns (specifying nanoseconds)
   in the timespec pointed by tvdestp. */
#define loadtspec(tvdestp,ns) do { \
  (tvdestp)->tv_sec = floor((ns) / NSEC_PER_SEC);	\
  (tvdestp)->tv_nsec = fmod((ns), NSEC_PER_SEC);	\
} while (0)

/* dtspec (tp) gives a double representing
   the timespec pointed by tp in seconds. */
#define dtspec(tp) ((tp)->tv_sec + (double)1e-9 * (tp)->tv_nsec)

extern char *ifname;			/* interface name */

extern void init_wireless_api(void);	/* init the wireless APIs */
extern iwrange *get_ranges();		/* contents ranges */
extern int get_channel_number();	/* frequency channel */
extern int get_bit_rate();		/* approximate bit rate in Mb/s */
extern int get_auto_bit_rate();		/* whether the bit rate is automatic */
extern int get_retry_limit(int flag);	/* short and long retry limits */
extern iwstats *get_stats();		/* contents of /proc/net/wireless */
extern void print_scan_info();          /* */ 

/*
 * Local Variables:
 * comment-column: 40
 * fill-column: 79
 * c-font-lock-extra-types: ("FILE"  "\\sw+_t" "bool" "iwrange" "iwstats")
 * End:
 */
